// Copyright (c) 2013 Vinay Bettadapura (Georgia Tech)

#include <algorithm>
#include <boost/lexical_cast.hpp>
#include <boost/regex.hpp>
#include <cmath>
#include <cstdio>
#include <fstream>
#include <iostream>
#include <limits>
#include <map>
#include <numeric>
#include <sstream>
#include <string>
#include <tbb/blocked_range.h>
#include <tbb/parallel_for.h>
#include <utility>
#include <vector>

#include "abow_core.h"
#include "definitions.h"
#include "reg_exp_gen_tbb.h"
#include "reg_exp_match_tbb.h"
#include "tf_df.h"
#include "utils.h"
#include "vector_space_similarity_tbb.h"

using std::cerr;
using std::cout;
using std::endl;
using std::ifstream;
using std::istringstream;
using std::map;
using std::ofstream;
using std::pair;
using std::set;
using std::string;
using std::vector;

namespace core {

ABOWCore::ABOWCore()
{
	// Default parameter values
	ngram_length_ = 3;
	no_of_time_bins_ = 4;
	scoring_metric_id_ = 5;
	time_precision_ = 2;
	duration_bins_ = "";
	regular_expressions_ = "";
	time_unit_ = "seconds";
	pyramid_ngrams_ = true;
	num_of_random_regexp_ = 0;
	similarity_precision_ = 0;
	print_zero_scores_ = true;

	input_event_file_ = "";
	output_histogram_data_file_ = ""; 
	output_similarity_scores_file_ = "";
	output_tfdf_histogram_data_file_ = "";
	output_global_similarity_score_file_ = "";
	output_global_term_frequency_file_ = "";
	output_word_document_frequency_file_ = "";
	data_source_ = "";
	no_of_tracks_ = 0;
	no_of_bins_ = 0;
	min_score_ = 0.0;
	max_score_ = 0.0;
	
	a_ngram_hist_device_ = 0;
	a_affinity_device_ = 0;
	a_affinity_bytes_ = 0;
	a_affinity_bytes_ = 0;
}

void ABOWCore::set_ngram_length(int n)
{
	ngram_length_ = n;
}

void ABOWCore::set_no_of_time_bins(int n)
{
	no_of_time_bins_ = n;
}

void ABOWCore::set_time_precision(int n)
{
	time_precision_ = n;
}

void ABOWCore::set_scoring_metric_id(int n)
{
	scoring_metric_id_ = n;
}

void ABOWCore::set_encoding(string s)
{
	if(s == "cumulative")
	{
		consider_cumulative_time_ = true;
		consider_interspersed_time_ = false;
	}
	else if(s == "interspersed")
	{
		consider_cumulative_time_ = false;
		consider_interspersed_time_ = true;
	}
	else
	{
		consider_cumulative_time_ = false;
		consider_interspersed_time_ = false;
	}
}

void ABOWCore::set_data_source(string s)
{
	data_source_ = s;
}

void ABOWCore::set_regular_expressions(string s)
{
	regular_expressions_ = s;
}

void ABOWCore::set_duration_bins(string s)
{
	duration_bins_ = s;
}

void ABOWCore::set_time_unit(string s)
{
	time_unit_ = s;
}

void ABOWCore::set_pyramid(string s)
{
	if(s == "yes")
		pyramid_ngrams_ = true;
	else
		pyramid_ngrams_ = false;
}

void ABOWCore::set_num_of_random_regexp(int n)
{
	num_of_random_regexp_ = n;
}

void ABOWCore::set_similarity_precision(int n)
{
	similarity_precision_ = n;
}

void ABOWCore::set_print_zero_scores(std::string s)
{
	if(s == "yes")
		print_zero_scores_ = true;
	else
		print_zero_scores_ = false;
}

bool ABOWCore::ConsiderTime() const
{
	return consider_cumulative_time_ || consider_interspersed_time_;
}

string ABOWCore::GetTrackID(const vector<string>& v_tokens) const
{
	return v_tokens[0];
}

string ABOWCore::GetEventID(const vector<string>& v_tokens) const
{
	return v_tokens[1];
}

string ABOWCore::GetStartTime(const vector<string>& v_tokens) const
{
	if(time_unit_ == "seconds")
		return v_tokens[4];
	else
		return v_tokens[2];
}

string ABOWCore::GetEndTime(const vector<string>& v_tokens) const
{
	if(time_unit_ == "seconds")
		return v_tokens[5];
	else
		return v_tokens[3];
}

void ABOWCore::SetPaths()
{
	input_event_file_ = data_source_ + "/events.txt";
	output_histogram_data_file_ =  data_source_ + "/op_histogram_data.txt";
	output_similarity_scores_file_ = data_source_ + "/op_similarity_scores.txt";
	output_tfdf_histogram_data_file_ = data_source_ +
										"/op_tfdf_histogram_data.txt";
	output_global_similarity_score_file_ = data_source_ +
										"/op_global_similarity_scores.txt";
	output_global_term_frequency_file_ = data_source_ +
										"/op_global_term_frequency.txt";
	output_word_document_frequency_file_ = data_source_ +
										"/op_word_document_frequency.txt";
}

void ABOWCore::BuildDurationRangeMapHelper(double start, double end, int index)
{
	pair<double, double> range;
	range.first = start;
	range.second = end;
	string str(1, static_cast<char>(65 + index));
	m_duration_range_map_[range] = str;
}

void ABOWCore::ComputeDurationBins()
{
	if(duration_bins_ == "")
	{
		double sum = 0.0;
		map<double, int>::const_iterator end = m_time_data_.end();
		for(map<double, int>::const_iterator iter = m_time_data_.begin();
			iter != end;
			iter++)
		{
			sum += iter->second;
		}

		double bin_capacity = round(sum/no_of_time_bins_);

		sum = 0;
		for(map<double, int>::const_iterator iter = m_time_data_.begin();
			iter != end;
			iter++)
		{
			if((sum == 0 && iter->second > bin_capacity) ||
			   (sum > bin_capacity))
			{
				if(duration_bins_ != "")
				{
					duration_bins_ += " ";
				}
				duration_bins_ += boost::lexical_cast<string>(iter->first);
				sum = 0;
			}
			else
			{
				sum += iter->second;
			}
		}
	}
}

void ABOWCore::BuildDurationRangeMap()
{
	ComputeDurationBins();

	string delimiter(" ");
	vector<string> time_bins;

	core::tokenize(time_bins, duration_bins_, delimiter);
	time_bins.push_back("LAST-DUMMY");

	for(vector<string>::size_type i = 0; i < time_bins.size() - 1; i++)
	{
		double start_frame = boost::lexical_cast<double>(time_bins[i]);
		
		double end_frame;
		// Replace LAST-DUMMY with Infinity
		if(i == time_bins.size() - 2)
		{
			end_frame = std::numeric_limits<double>::infinity();
		}
		else
		{
			end_frame = boost::lexical_cast<double>(time_bins[i + 1]);
		}

		pair<double, double> range;
		if(i == 0)
		{
			BuildDurationRangeMapHelper(0, start_frame, i+0);
			BuildDurationRangeMapHelper(start_frame, end_frame, i+1);
		}
		else
		{
			BuildDurationRangeMapHelper(start_frame, end_frame, i+1);
		}
	}
}

int ABOWCore::AugmentEventsWithTimeInfo(const vector<string>& v_tokens,
									int event_position,
									vector<string>& v_event_data_single_track)
{
	if(event_position == EVENT_POSITION::FIRST_IN_TRACK)
	{
		v_event_data_single_track.clear();
	}

	if(event_position == EVENT_POSITION::INTERMEDIATE)
	{
		double diff = boost::lexical_cast<double>(GetStartTime(v_tokens)) -
			boost::lexical_cast<double>(
			v_event_data_single_track[v_event_data_single_track.size() - 2]);

		if(diff < 0)
		{
			cerr << "ERROR IN AugmentEventsWithTimeInfo(): time diff is < 0\n";
			return -1;
		}
		v_event_data_single_track.push_back(boost::lexical_cast<string>(diff));
		v_event_data_single_track.push_back("D");
		v_event_data_single_track.push_back("D");

		if(boost::lexical_cast<int>(GetTrackID(v_tokens)) >= 0)
		{
			diff = boost::lexical_cast<double>(round(diff *
											pow(10.0, time_precision_))) /
											pow(10.0, time_precision_);
			++m_time_data_[diff];
		}
	}

	v_event_data_single_track.push_back(GetEventID(v_tokens));
	v_event_data_single_track.push_back(GetStartTime(v_tokens));
	v_event_data_single_track.push_back(GetEndTime(v_tokens));

	return 0;
}

int ABOWCore::BuildEventData()
{
	cout << "\nReading activities (event streams) from: "
		<< input_event_file_ << "\n";
	ifstream ifs_input(input_event_file_.c_str());

	string str_line;
	vector<string> v_tokens;
	vector<string> v_event_data_single_track;
	string delimiter(" ");

	int curr_track_id = 0, previous_track_id = 0, line_no = 1;

	if(ifs_input.is_open())
	{
		while(getline(ifs_input, str_line))
		{
			v_tokens.clear();
			core::tokenize(v_tokens, str_line, delimiter);

			if(v_tokens.size() != GLOBAL_CONSTANTS::NUMBER_OF_EVENT_FIELDS)
			{
				cerr << "ERROR IN BuildEventData(): Event does not have the "
						"right number of fields\n"; 
				return -1;
			}
			
			if(line_no == 1)
			{
				previous_track_id = boost::lexical_cast<int>(
													GetTrackID(v_tokens));
				AugmentEventsWithTimeInfo(v_tokens,
										EVENT_POSITION::FIRST_IN_FILE,
										v_event_data_single_track);
				line_no++;
			}
			else
			{
				curr_track_id = boost::lexical_cast<int>(GetTrackID(v_tokens));
				if(curr_track_id != previous_track_id)
				{
					vv_event_data_all_tracks_.push_back(
												v_event_data_single_track);
					v_track_ids_.push_back(previous_track_id);
					AugmentEventsWithTimeInfo(v_tokens,
											EVENT_POSITION::FIRST_IN_TRACK,
											v_event_data_single_track);
				}
				else
				{
					int ret = AugmentEventsWithTimeInfo(v_tokens,
												EVENT_POSITION::INTERMEDIATE,
												v_event_data_single_track);
					if(ret == -1)
					{
						cerr << "ERROR IN BuildEventData(): Augmenting events "
								"with time failed\n"; 
						return -1;
					}
				}
				previous_track_id = curr_track_id;
			}
		}
		vv_event_data_all_tracks_.push_back(v_event_data_single_track);
		v_track_ids_.push_back(curr_track_id);
		ifs_input.close();
	}
	else
	{
		cerr << "ERROR IN BuildEventData(): Failed to open event file\n";
		return -1;
	}

	no_of_tracks_ = vv_event_data_all_tracks_.size();
	
	if(no_of_tracks_ != v_track_ids_.size())
	{
		cerr << "ERROR IN BuildEventData(): Mismatch between number of tracks "
				"and track ids\n"; 
		return -1;
	}

	return 0;
}

void ABOWCore::PrintDurationInfo() const
{
	if(ConsiderTime())
	{
		cout << "\nCreated time bins:\n";

		map<pair<double, double>, string>::const_iterator end =
							m_duration_range_map_.end();
		for(map<pair<double, double>, string>::const_iterator iter =
							m_duration_range_map_.begin(); iter != end; iter++)
		{
			cout << iter->second << ": From " << iter->first.first
				<< " to " << iter->first.second << "\n";
		}
	}
}

string ABOWCore::FindDurationCategoryByRange(double range) const
{
	string result = "ERROR";
	map<pair<double, double>, string>::const_iterator end =
							m_duration_range_map_.end();
	for(map<pair<double, double>, string>::const_iterator iter =
							m_duration_range_map_.begin(); iter != end; iter++)
	{
		if(range >= iter->first.first && range <= iter->first.second)
		{
			result = iter->second;
			break;
		}
	}
	if(result == "ERROR")
	{
		cerr << "\nError in event.txt file. Run your event file through "
				"event_file_validator.rb to check for errors. Aborting!\n\n";
		exit(-1);
	}
	return result;
}

void ABOWCore::BinDurations()
{
	PrintDurationInfo();

	for(vector<string>::size_type i = 0;
		i < vv_event_data_all_tracks_.size(); ++i)
	{
		for(vector<string>::size_type j = 3;
			j < vv_event_data_all_tracks_[i].size(); j += 6)
		{
			double range = boost::lexical_cast<double>(
											vv_event_data_all_tracks_[i][j]);
			vv_event_data_all_tracks_[i][j] =
							FindDurationCategoryByRange(range);
		}
	}
}

void ABOWCore::BuildEventStrings()
{
	for(vector<string>::size_type i = 0;
		i < vv_event_data_all_tracks_.size(); ++i)
	{
		string str = "";
		for(vector<string>::size_type j = 0;
			j < vv_event_data_all_tracks_[i].size(); ++j)
		{
			if(j % 3 == 0)
			{
				str += vv_event_data_all_tracks_[i][j];

				s_all_symbols_.insert(vv_event_data_all_tracks_[i][j]);
				
				++m_symbol_frequency_[vv_event_data_all_tracks_[i][j]];
			}
		}
		v_event_strings_.push_back(str);
	}
}

void ABOWCore::BuildAllSymbols()
{
	std::copy(s_all_symbols_.begin(),
			s_all_symbols_.end(),
			std::back_inserter(v_all_symbols_));
}

void ABOWCore::BuildPdfCdf()
{
	double sum = 0.0;
	for(vector<string>::size_type i = 0; i < v_all_symbols_.size(); ++i)
	{
		sum += m_symbol_frequency_[v_all_symbols_[i]];
	}

	for(vector<string>::size_type i = 0; i < v_all_symbols_.size(); ++i)
	{
		double prob = m_symbol_frequency_[v_all_symbols_[i]]/sum;
		v_symbol_pdf_.push_back(prob);

	}

	v_symbol_cdf_.push_back(v_symbol_pdf_[0]);
	for(vector<string>::size_type i = 1; i < v_symbol_pdf_.size(); ++i)
	{
		double cumulative = v_symbol_pdf_[i] + v_symbol_cdf_[i-1];
		v_symbol_cdf_.push_back(cumulative);
	}
}

void ABOWCore::BuildTimeBins()
{
	map<pair<double, double>, string>::const_iterator iter =
												m_duration_range_map_.begin();
	map<pair<double, double>, string>::const_iterator end =
												m_duration_range_map_.end();
	for(; iter != end; iter++)
	{
		v_time_symbols_.push_back(iter->second);
	}
}

void ABOWCore::BuildObservableSymbols()
{
	std::set_difference(v_all_symbols_.begin(),
		v_all_symbols_.end(), 
		v_time_symbols_.begin(),
		v_time_symbols_.end(),
		std::back_inserter(v_observable_events_));
}

void ABOWCore::BuildRandomRegExpBins()
{
	if(num_of_random_regexp_ == 0)
		return;

	BuildEventStrings();
	BuildAllSymbols();
	BuildPdfCdf();

	v_random_reg_exps_.resize(num_of_random_regexp_);

	cout << "\nGenerating " << num_of_random_regexp_
		<< " random regular expressions..." << endl;
	tbb::parallel_for(
		tbb::blocked_range<int>(
			0,
			num_of_random_regexp_
		), 
		RegExpGenTBB(
			&v_random_reg_exps_,
			&v_event_strings_,
			&s_generated_regexps_,
			&v_symbol_cdf_,
			&v_all_symbols_,
			v_all_symbols_,
			v_all_symbols_.size()
			//10
		)
	);
	cout << "Done.\n";

	vector<string> tmp;
	v_rand_regex_.resize(v_event_strings_.size(), tmp);
	
	cout << "\nMatching the generated regular expressions..." << endl;
	tbb::parallel_for(tbb::blocked_range<unsigned long>(
		0, v_event_strings_.size()), 
		RegExpMatchTBB(&v_random_reg_exps_, &v_event_strings_, &v_rand_regex_));
	cout << "Done.\n";
}

void ABOWCore::SetIndexParameters()
{
	if(consider_cumulative_time_)
	{
		index_jump_ = 6;
		index_tweak_ = 3;
	}
	else if(consider_interspersed_time_)
	{
		index_jump_ = 3;
		index_tweak_ = 0;
	}
	else
	{
		index_jump_ = 6;
		index_tweak_ = 3;
	}
}

void ABOWCore::BuildNGramsHelper(int i, int length_of_ngram,
								vector<string>& v_ngram)
{
	vector<string>::size_type size = vv_event_data_all_tracks_[i].size();

	if(size == 3)
	{
		string str_ngram = "";
		if(ConsiderTime())
		{
			if(consider_cumulative_time_)
			{
				double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][2]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
				str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
							FindDurationCategoryByRange(diff);
			}
			if(consider_interspersed_time_)
			{
				str_ngram = vv_event_data_all_tracks_[i][0];
			}
		}
		else
		{
			str_ngram = vv_event_data_all_tracks_[i][0];
		}
		v_ngram.push_back(str_ngram);
	}
	else if(size == 9)
	{
		string str_ngram = "";
		if(length_of_ngram == 1)
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
						boost::lexical_cast<double>(vv_event_data_all_tracks_[i][2])
						- boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);

					diff = boost::lexical_cast<double>(vv_event_data_all_tracks_[i][8])
						 - boost::lexical_cast<double>(vv_event_data_all_tracks_[i][7]);
					str_ngram = vv_event_data_all_tracks_[i][6] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][3]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][6]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0]);
				v_ngram.push_back(vv_event_data_all_tracks_[i][6]);
			}
		}
		else if(length_of_ngram == 2)
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][8]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
									vv_event_data_all_tracks_[i][3]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6]);
			}
		}
		else
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][8]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								FindDurationCategoryByRange(diff);
				}
				if(consider_interspersed_time_)
				{
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][3] + "|" +
								vv_event_data_all_tracks_[i][6];
				}
			}
			else
			{
				str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
							vv_event_data_all_tracks_[i][6];
			}
			v_ngram.push_back(str_ngram);
		}
	}
	else if(size == 15)
	{
		string str_ngram = "";
		if(length_of_ngram == 1)
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][2]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);

					diff = boost::lexical_cast<double>(vv_event_data_all_tracks_[i][8])
						 - boost::lexical_cast<double>(vv_event_data_all_tracks_[i][7]);
					str_ngram = vv_event_data_all_tracks_[i][6] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);

					diff = boost::lexical_cast<double>(vv_event_data_all_tracks_[i][14])
						 - boost::lexical_cast<double>(vv_event_data_all_tracks_[i][13]);
					str_ngram = vv_event_data_all_tracks_[i][12] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][3]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][6]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][9]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][12]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0]);
				v_ngram.push_back(vv_event_data_all_tracks_[i][6]);
				v_ngram.push_back(vv_event_data_all_tracks_[i][12]);
			}
		}
		else if(length_of_ngram == 2)
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][8]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);

					diff = boost::lexical_cast<double>(vv_event_data_all_tracks_[i][14])
						 - boost::lexical_cast<double>(vv_event_data_all_tracks_[i][7]);
					str_ngram = vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
									vv_event_data_all_tracks_[i][3]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][6] + "|" +
									vv_event_data_all_tracks_[i][9]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][9] + "|" +
									vv_event_data_all_tracks_[i][12]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6]);
				v_ngram.push_back(vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12]);
			}
		}
		else if(length_of_ngram == 3)
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][14]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
									vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6] + "|" +
									vv_event_data_all_tracks_[i][9]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][6] + "|" +
									vv_event_data_all_tracks_[i][9] + "|" +
									vv_event_data_all_tracks_[i][12]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12]);
			}
		}
		else if(length_of_ngram == 4)
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][14]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
									vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6] + "|" +
									vv_event_data_all_tracks_[i][9]);
					v_ngram.push_back(vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6] + "|" +
									vv_event_data_all_tracks_[i][9] + "|" +
									vv_event_data_all_tracks_[i][12]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12]);
			}
		}
		else
		{
			if(ConsiderTime())
			{
				if(consider_cumulative_time_)
				{
					double diff =
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][14]) -
					boost::lexical_cast<double>(vv_event_data_all_tracks_[i][1]);
					str_ngram = vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12] + "|" +
								FindDurationCategoryByRange(diff);
					v_ngram.push_back(str_ngram);
				}
				if(consider_interspersed_time_)
				{
					v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
									vv_event_data_all_tracks_[i][3] + "|" +
									vv_event_data_all_tracks_[i][6] + "|" +
									vv_event_data_all_tracks_[i][9] + "|" +
									vv_event_data_all_tracks_[i][12]);
				}
			}
			else
			{
				v_ngram.push_back(vv_event_data_all_tracks_[i][0] + "|" +
								vv_event_data_all_tracks_[i][6] + "|" +
								vv_event_data_all_tracks_[i][12]);
			}
		}
	}
	else
	{
		for(vector<string>::size_type j = 0; j < size; j += index_jump_)
		{
			if(j > size - length_of_ngram * index_jump_ + index_tweak_)
			{
				break;
			}
			string str_ngram = "";
			vector<string>::size_type k = 0;
			for(k = j;
				k < size && k < j + length_of_ngram * index_jump_;
				k += index_jump_)
			{
				str_ngram = str_ngram + vv_event_data_all_tracks_[i][k] + "|";
			}
			if(consider_cumulative_time_)
			{
				double diff =
				boost::lexical_cast<double>(vv_event_data_all_tracks_[i][k - 4]) -
				boost::lexical_cast<double>(vv_event_data_all_tracks_[i][j + 1]);
				str_ngram = str_ngram + FindDurationCategoryByRange(diff) + "|";
			}
			str_ngram = str_ngram.substr(0, str_ngram.length() - 1);
			v_ngram.push_back(str_ngram);
		}
	}
}

void ABOWCore::BuildNGrams()
{
	SetIndexParameters();

	for(vector<string>::size_type i = 0;
		i < vv_event_data_all_tracks_.size();
		i++)
	{
		vector<string> v_ngram;
		if(pyramid_ngrams_)
		{
			for(int j = 1; j <= ngram_length_; ++j)
			{
				BuildNGramsHelper(i, j, v_ngram);
			}
		}
		else
		{
			BuildNGramsHelper(i, ngram_length_, v_ngram);
		}
		v_ngrams_.push_back(v_ngram);
	}
}

void ABOWCore::BuildRegExps()
{
	string delimiter(" ");
	v_regular_expressions_.clear();
	core::tokenize(v_regular_expressions_, regular_expressions_, delimiter);
}

void ABOWCore::BuildNGramsAllHistogram()
{
	for(vector<vector<string> >::size_type i = 0; i < v_ngrams_.size(); ++i)
	{
		for(vector<string>::size_type j = 0; j < v_ngrams_[i].size(); ++j)
		{
			for(vector<string>::size_type k = 0;
				k < v_regular_expressions_.size();
				++k)
			{
				boost::regex reg_exp(v_regular_expressions_[k]);
				if(boost::regex_match(v_ngrams_[i][j], reg_exp))
				{
					++m_ngram_all_histogram_[v_regular_expressions_[k]];
					s_regexp_matches_.insert(v_ngrams_[i][j]);
				}
			}
			++m_ngram_all_histogram_[v_ngrams_[i][j]];
		}
	}

	if(num_of_random_regexp_ > 0)
	{
		for(vector<vector<string> >::size_type i = 0;
			i < v_rand_regex_.size();
			++i)
		{
			for(vector<string>::size_type j = 0;
				j < v_rand_regex_[i].size();
				++j)
			{
				++m_ngram_all_histogram_[v_rand_regex_[i][j]];
			}
		}
	}
}

void ABOWCore::BuildDiscoveredWords()
{
	ofstream ofs_global_tf(output_global_term_frequency_file_.c_str());
	
	vector<string>::size_type index = 0;
	map<string, int>::const_iterator end = m_ngram_all_histogram_.end();
	for(map<string, int>::const_iterator iter = m_ngram_all_histogram_.begin();
		iter != end; iter++)
	{
		v_discovered_words_.push_back(iter->first);

		m_ngrams_index_[iter->first] = index;
		index++;
		
		ofs_global_tf << iter->first << " " << iter->second << "\n";
	}
	no_of_bins_ = v_discovered_words_.size();
	
	cout << "\nGlobal term-frequency (TF) scores written to: "
		<< output_global_term_frequency_file_ << "\n";
	ofs_global_tf.close();
}

void ABOWCore::WriteSimilarityFile() const
{
	ofstream ofs_similarity(output_similarity_scores_file_.c_str());
	map<double, vector<int> > m_temp_matches;

	if(similarity_precision_ > 0)
	{
		ofs_similarity.precision(similarity_precision_);
	}

	for(unsigned long long i = 0; i < no_of_tracks_; i++)
	{
		ofs_similarity << v_track_ids_[i] << " ";
		m_temp_matches.clear();
		for(unsigned long long j = 0; j < no_of_tracks_; j++)
		{
			unsigned long long affinity_idx = i * no_of_tracks_ + j;
			if(!print_zero_scores_ && a_affinity_[affinity_idx] == 0)
			{
				continue;
			}
			m_temp_matches[a_affinity_[affinity_idx]].push_back(v_track_ids_[j]);
		}

		map<double, vector<int> >::const_reverse_iterator end =
														m_temp_matches.rend();
		for(map<double, vector<int> >::const_reverse_iterator iter =
			m_temp_matches.rbegin(); iter != end; iter++)
		{
			vector<int>::const_iterator vect_end = iter->second.end();
			for(vector<int>::const_iterator vect_itr = iter->second.begin();
				vect_itr != vect_end; vect_itr++)
			{
				ofs_similarity << *vect_itr << ":" << iter->first << " ";
			}
		}
		ofs_similarity << "\n";
	}
	ofs_similarity.close();
	cout << "\nSimilarity scores written to: "
		<< output_similarity_scores_file_ << "\n";
}

void ABOWCore::ComputeMinMaxScore()
{
	min_score_ = std::numeric_limits<double>::infinity();
	max_score_ = -std::numeric_limits<double>::infinity();

	for(vector<double>::size_type i = 0; i < v_scores_.size(); i++)
	{
		if(v_scores_[i] < min_score_)
		{
			min_score_ = v_scores_[i];
		}
		if(v_scores_[i] > max_score_)
		{
			max_score_ = v_scores_[i];
		}
	}
}

void ABOWCore::WriteGlobalSimilarityScoreFile() const
{
	ofstream ofs_score(output_global_similarity_score_file_.c_str());

	for(vector<double>::size_type i = 0; i < v_scores_.size(); i++)
	{
		ofs_score << v_track_ids_[i] << " " 
				<< (v_scores_[i]-min_score_)/(max_score_-min_score_) << "\n";
	}

	ofs_score.close();
	cout << "\nGlobal similarity scores written to: "
		<< output_global_similarity_score_file_ << "\n\n";
}

int ABOWCore::AllocateMemoryForAffinityMarix()
{
	unsigned long long no_of_array_elements = no_of_tracks_ * no_of_tracks_;
	a_affinity_bytes_ = no_of_array_elements * sizeof(float);

	cout << "\nBuiding affinity matrix: Allocating memory for " << no_of_tracks_
		<< " * " << no_of_tracks_ << " = " << no_of_array_elements
		<< " elements ("  << a_affinity_bytes_/1024.0 << " KB, "
		<< a_affinity_bytes_/1024.0/1024.0 << " MB)\n";

	float* affinitystorage = new float[no_of_array_elements];
	if(affinitystorage == NULL)
	{
		cout << "Memory allocation failed!\n";
		return -1;
	}
	
	memset(affinitystorage, 0, a_affinity_bytes_);
	
	a_affinity_ = boost::shared_array<float>(affinitystorage);

	if(a_affinity_ == 0)
	{
		cout << "Memory allocation failed!\n";
		return -1;
	}

	cout << "Successfully allocated memory\n";

	return 0;
}

#if USE_CUDA
extern int VectorSpaceSimilarityCUDA(float* a_ngram_hist_,
									float* a_ngram_hist_device_,
									unsigned long a_ngram_hist_bytes_,
									float* a_affinity_,
									float* a_affinity_device_,
									unsigned long a_affinity_bytes_,
									size_t no_of_tracks_,
									size_t no_of_bins_);
#endif

int ABOWCore::BuildAffinityAndScores()
{
	v_scores_.resize(no_of_tracks_);

#if USE_CUDA
	VectorSpaceSimilarityCUDA(a_ngram_hist_,
							a_ngram_hist_device_,
							a_ngram_hist_bytes_,
							a_affinity_,
							a_affinity_device_,
							a_affinity_bytes_,
							no_of_tracks_,
							no_of_bins_);
#else
	tbb::parallel_for(tbb::blocked_range<unsigned long long>(0, no_of_tracks_), 
		VectorSpaceSimilarityTBB(a_ngram_hist_,
								a_affinity_,
								no_of_tracks_,
								no_of_bins_));
#endif

	for(unsigned long long i = 0; i < no_of_tracks_; i++)
	{
		v_scores_[i] = std::accumulate(a_affinity_.get() + i*no_of_tracks_,
				a_affinity_.get() + i*no_of_tracks_ + no_of_tracks_, 0.0) - 1;
	}

	return 0;
}

void ABOWCore::ComputeTfIdf()
{
	TfDf tf_df_reweighter;

	for(vector<string>::size_type i = 0; i < no_of_tracks_; ++i)
	{
		tf_df_reweighter.reweight(i, this);
	}
}

void ABOWCore::WriteDocumentFrequencyFile() const
{
  	ofstream ofs_word_df(output_word_document_frequency_file_.c_str());

	map<string, int>::const_iterator end = m_ngram_hist_across_.end();
	for(map<string, int>::const_iterator iter = m_ngram_hist_across_.begin();
		iter != end; iter++)
	{
		ofs_word_df << iter->first << " " << iter->second << "\n";
	}

	cout << "\nWord document-frequency (DF) scores written to: "
		<< output_word_document_frequency_file_ << "\n";
	ofs_word_df.close();
}

void ABOWCore::BuildNGramHistAcross(vector<string>::size_type i)
{
	vector<string>::size_type begin = no_of_bins_ * i;
	for(vector<string>::size_type iter = 0; iter < no_of_bins_; ++iter)
	{
		if(a_ngram_hist_[begin+iter] > 0)
		{
			++m_ngram_hist_across_[v_discovered_words_[iter]];
		}
	}
}

void ABOWCore::WriteHistDataToFile(vector<string>::size_type i,
								ofstream& ofs_data)
{
	ofs_data << v_track_ids_[i] << " ";

	vector<string>::size_type begin = no_of_bins_ * i;
	for(vector<string>::size_type iter = 0; iter < no_of_bins_; ++iter)
	{
		ofs_data << v_discovered_words_[iter] << ":"
				<< a_ngram_hist_[begin+iter] << " ";
	}
	
	ofs_data << "\n";
}

int ABOWCore::AllocateMemoryForNGramHists()
{
	size_t no_of_hist_elements = no_of_bins_ * no_of_tracks_;
	a_ngram_hist_bytes_ = no_of_hist_elements * sizeof(float);

	cout << "\nBuilding augmented bag-of-words: Allocating memory for "
	  	<< no_of_tracks_ << " * " << no_of_bins_ << " = "
		<< no_of_hist_elements << " elements ("
		<< a_ngram_hist_bytes_/1024.0 << " KB, "
		<< a_ngram_hist_bytes_/1024.0/1024.0 << " MB)...\n";

	float* histstorage = new float[no_of_hist_elements];
	if(histstorage == NULL)
	{
		cout << "Memory allocation failed!\n";
		return -1;
	}

	memset(histstorage, 0, a_ngram_hist_bytes_);

	a_ngram_hist_ = boost::shared_array<float>(histstorage);

	if(a_ngram_hist_ == 0)
	{
		cout << "Memory allocation failed!\n";
		return -1;
	}

	cout << "Successfully allocated memory\n";

	return 0;
}

void ABOWCore::BuildNGramHists()
{
	ofstream ofs_data(output_histogram_data_file_.c_str());

	for(vector<string>::size_type i = 0; i < no_of_tracks_; ++i)
	{
		vector<string>::size_type index;

		for(vector<string>::size_type j = 0; j < v_ngrams_[i].size(); ++j)
		{
			for(vector<string>::size_type k = 0;
				k < v_regular_expressions_.size(); ++k)
			{
				boost::regex reg_exp(v_regular_expressions_[k]);
				if(boost::regex_match(v_ngrams_[i][j], reg_exp))
				{
					index = i * no_of_bins_ +
							m_ngrams_index_[v_regular_expressions_[k]];
					a_ngram_hist_[index] = a_ngram_hist_[index] + 1;
				}
			}

			index = i * no_of_bins_ + m_ngrams_index_[v_ngrams_[i][j]];
			a_ngram_hist_[index] = a_ngram_hist_[index] + 1;
		}

		if(num_of_random_regexp_ > 0)
		{
			for(vector<string>::size_type j = 0;
				j < v_rand_regex_[i].size(); ++j)
			{
				index = i * no_of_bins_ + m_ngrams_index_[v_rand_regex_[i][j]];
				a_ngram_hist_[index] = a_ngram_hist_[index] + 1;
			}
		}

		WriteHistDataToFile(i, ofs_data);
		BuildNGramHistAcross(i);
	}
	
	cout << "\nHistogram data written to: "
		<< output_histogram_data_file_ << "\n";
	ofs_data.close();
}

int ABOWCore::WriteTfIdfHistogramScores()
{
	ofstream ofs_output_col(output_tfdf_histogram_data_file_.c_str());

	for(vector<string>::size_type i = 0; i < no_of_tracks_; ++i)
	{
		ofs_output_col << v_track_ids_[i] << " ";

		vector<string>::size_type begin = no_of_bins_ * i;
		for(vector<string>::size_type iter = 0; iter < no_of_bins_; ++iter)
		{
			ofs_output_col << v_discovered_words_[iter] << ":"
						<< a_ngram_hist_[begin+iter] << " ";
		}
		ofs_output_col << "\n";
	}

	ofs_output_col.close();

	cout << "\nTF-DF weighted histogram scores written to: "
		<< output_tfdf_histogram_data_file_ << "\n";

	return 0;
}

int ABOWCore::Compute()
{
	SetPaths();
	if(BuildEventData() < 0) return -1;
	BuildDurationRangeMap();
	BinDurations();
	BuildNGrams();
	BuildRegExps();
	BuildRandomRegExpBins();
	BuildNGramsAllHistogram();
	BuildDiscoveredWords();
	if(AllocateMemoryForNGramHists() < 0) return -1;
	BuildNGramHists();
	WriteDocumentFrequencyFile();
	ComputeTfIdf();
	if(WriteTfIdfHistogramScores() < 0) return -1;
	if(AllocateMemoryForAffinityMarix() < 0) return -1;
	if(BuildAffinityAndScores() < 0) return -1;
	WriteSimilarityFile();
	ComputeMinMaxScore();
	WriteGlobalSimilarityScoreFile();
	
	return 0;
}

}
