// Copyright (c) 2013 Vinay Bettadapura (Georgia Tech)

#ifndef _CORE_ABOW_CORE_H_
#define _CORE_ABOW_CORE_H_

#include <boost/shared_array.hpp>
#include <map>
#include <set>
#include <string>
#include <vector>

#include "config.h"
#include "tf_df.h"

namespace core {

class ABOWCore
{
	friend class TfDf;

public:
	ABOWCore();

	void set_data_source(std::string s);
	void set_ngram_length(int n);
	void set_no_of_time_bins(int n);
	void set_scoring_metric_id(int n);
	void set_encoding(std::string s);
	void set_time_precision(int n);
	void set_min_no_of_data_points(int i);
	void set_regular_expressions(std::string s);
	void set_duration_bins(std::string s);
	void set_time_unit(std::string s);
	void set_pyramid(std::string s);
	void set_num_of_random_regexp(int n);
	void set_similarity_precision(int n);
	void set_print_zero_scores(std::string s);

	int Compute();
	int CreateTrainingFiles();

private:
	void SetPaths();
	int BuildEventData();
	void BinDurations();
	bool ConsiderTime() const;
	std::string GetTrackID(const std::vector<std::string>& v_tokens) const;
	std::string GetEventID(const std::vector<std::string>& v_tokens) const;
	std::string GetStartFrame(const std::vector<std::string>& v_tokens) const;
	std::string GetEndFrame(const std::vector<std::string>& v_tokens) const;
	std::string GetStartTime(const std::vector<std::string>& v_tokens) const;
	std::string GetEndTime(const std::vector<std::string>& v_tokens) const;
	int AugmentEventsWithTimeInfo(const std::vector<std::string>& v_tokens,
						int event_position,
						std::vector<std::string>& v_event_data_single_track); 
	void BuildDurationRangeMap();
	void BuildDurationRangeMapHelper(double start, double end, int index);
	void PrintDurationInfo() const;
	std::string FindDurationCategoryByRange(double range) const;
	void BuildNGrams();
	void BuildNGramsHelper(int i, int length_of_ngram,
						std::vector<std::string>& v_ngram);
	void BuildNGramsAllHistogram();
	void BuildEventStrings();
	void BuildAllSymbols();
	void BuildPdfCdf();
	void BuildObservableSymbols();
	void BuildTimeBins();
	void BuildRandomRegExpBins();
	void SetIndexParameters();
	void BuildDiscoveredWords();
	void WriteSimilarityFile() const;
	int AllocateMemoryForAffinityMarix();
	int BuildAffinityAndScores();
#if USE_CUDA
	int BuildAffinityAndScoresCUDA();
#endif
	void ComputeMinMaxScore();
	void WriteGlobalSimilarityScoreFile() const;
	int WriteTfIdfHistogramScores();
	void PrintScoreDetails() const;
	void ComputeTfIdf();
	void BuildNGramHistAcross(std::vector<std::string>::size_type i);
	void WriteDocumentFrequencyFile() const;
	void WriteHistDataToFile(std::vector<std::string>::size_type i,
							std::ofstream& ofs_data);
	int AllocateMemoryForNGramHists();
	void BuildNGramHists();
	void ComputeDurationBins();
	void PrintNGramHistForAllTracks() const;
	void BuildRegExps();

	int ngram_length_;
	int no_of_time_bins_;
	int scoring_metric_id_;
	int time_precision_;
	double min_score_;
	double max_score_;
	bool consider_cumulative_time_;
	bool consider_interspersed_time_;
	bool consider_time_;
	bool calculate_in_percentages_;
	bool calculate_in_seconds_;
	std::string duration_bins_;
	std::string input_event_file_;
	std::string data_source_;
	std::string output_histogram_data_file_;
	std::string output_similarity_scores_file_;
	std::string output_global_similarity_score_file_;
	std::string output_tfdf_histogram_data_file_;
	std::string output_global_term_frequency_file_;
	std::string	output_word_document_frequency_file_;
	std::string spatial_;
	std::string regular_expressions_;
	std::string time_unit_;
	bool pyramid_ngrams_;
	int num_of_random_regexp_;
	int similarity_precision_;
	bool print_zero_scores_;

	std::vector<std::string>::size_type index_jump_;
	std::vector<std::string>::size_type index_tweak_;
	unsigned long long no_of_tracks_;
	std::vector<std::string>::size_type no_of_bins_;
	boost::shared_array<float> a_affinity_;
	float* a_affinity_device_;
	std::vector<double> v_scores_;
	std::vector<int> v_track_ids_;
	unsigned long long a_affinity_bytes_;

	std::map<std::pair<double, double>, std::string> m_duration_range_map_;
	std::vector<std::vector<std::string> > vv_event_data_all_tracks_;
	std::vector<std::vector<std::string> > v_ngrams_;
	std::map<std::string, int> m_ngram_all_histogram_;
	std::map<double, int> m_time_data_;
	std::vector<std::string> v_discovered_words_;
	std::map<std::string, std::vector<std::string>::size_type> m_ngrams_index_;
	boost::shared_array<float> a_ngram_hist_;
	float* a_ngram_hist_device_;
	std::map<std::string, int> m_ngram_hist_across_;
	std::vector<std::string> v_regular_expressions_;
	unsigned long a_ngram_hist_bytes_;
	std::set<std::string> s_regexp_matches_;
	std::vector<std::vector<std::string> > v_rand_regex_;
	std::vector<std::string> v_event_strings_;
	std::set<std::string> s_all_symbols_;
	std::vector<std::string> v_all_symbols_;
	std::map<std::string, double> m_symbol_frequency_;
	std::vector<double> v_symbol_pdf_;
	std::vector<double> v_symbol_cdf_;
	std::vector<std::string> v_time_symbols_;
	std::vector<std::string> v_observable_events_;
	std::vector<std::string> v_random_reg_exps_;
	std::set<std::string> s_generated_regexps_;
};

}

#endif
