// Copyright (c) 2013 Vinay Bettadapura (Georgia Tech)

#include <cmath>
#include <map>
#include <string>
#include <vector>

#include "tf_df.h"

using std::vector;
using std::string;
using std::map;

namespace core {

void TfDf::reweight(vector<string>::size_type ti, ABOWCore* vc)
{
	track_index = ti;
	abow_core = vc;

	if(vc->scoring_metric_id_ == 1)
		natural_no();
	else if(vc->scoring_metric_id_ == 2)
		natural_idf();
	else if(vc->scoring_metric_id_ == 3)
		natural_probidf();
	else if(vc->scoring_metric_id_ == 4)
		log_no();
	else if(vc->scoring_metric_id_ == 5)
		log_idf();
	else if(vc->scoring_metric_id_ == 6)
		log_probidf();
	else if(vc->scoring_metric_id_ == 7)
		augmented_no();
	else if(vc->scoring_metric_id_ == 8)
		augmented_idf();
	else if(vc->scoring_metric_id_ == 9)
		augmented_probidf();
	else if(vc->scoring_metric_id_ == 10)
		boolean_no();
	else if(vc->scoring_metric_id_ == 11)
		boolean_idf();
	else if(vc->scoring_metric_id_ == 12)
		boolean_probidf();
	else if(vc->scoring_metric_id_ == 13)
		logavg_no();
	else if(vc->scoring_metric_id_ == 14)
		logavg_idf();
	else if(vc->scoring_metric_id_ == 15)
		logavg_probidf();
}

/*******************************************************************************
 * General terms:
 * tf : Term frequency
 * df : Document frequency
 * N  : Number of documents (i.e. activities, in our case)
 * tf_max : Maximum term frequency
 ******************************************************************************/

/*******************************************************************************
 * Method 1: 
 * N (Natural) : tf
 * N (No)      : 1
 ******************************************************************************/
void TfDf::natural_no()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = 1;

		double TF = tf;
		double DF = df;

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 2:
 * N (Natural) : tf
 * T (IDF)     : log(N/df)
 ******************************************************************************/
void TfDf::natural_idf()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = tf;
		double DF = log(abow_core->no_of_tracks_/df);

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 3:
 * N (Natural)  : tf
 * P (Prob IDF) : max{0, log[(N - df)/df]}
 ******************************************************************************/
void TfDf::natural_probidf()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = tf;
		double DF = std::max(0.0, log((abow_core->no_of_tracks_ - df)/df));

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 4:
 * L (Logarithm) : 1 + log(tf) if tf > 0, 0 otherwise
 * N (No)        : 1
 ******************************************************************************/
void TfDf::log_no()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];

		double TF = 0.0; if(tf > 0) { TF = 1 + log(tf); }
		double DF = 1;

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 5:
 * L (Logarithm) : 1 + log(tf) if tf > 0, 0 otherwise
 * T (IDF)       : log(N/df)
 ******************************************************************************/
void TfDf::log_idf()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = 0.0; if(tf > 0) { TF = 1 + log(tf); }
		double DF = log(abow_core->no_of_tracks_/df);

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 6:
 * L (Logarithm) : 1 + log(tf) if tf > 0, 0 otherwise
 * P (Prob IDF)  : max{0, log[(N - df)/df]}
 ******************************************************************************/
void TfDf::log_probidf()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = 0.0; if(tf > 0) { TF = 1 + log(tf); }
		double DF = std::max(0.0, log((abow_core->no_of_tracks_ - df)/df));

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 7:
 * A (Augmented) : a + (1 - a)(tf/tf_max)
 * N (No)        : 1
 ******************************************************************************/
void TfDf::augmented_no()
{
	double tf_max = -99999;

	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		if(abow_core->a_ngram_hist_[begin+iter] > tf_max)
			tf_max = abow_core->a_ngram_hist_[begin+iter];
	}

	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double a = 0.5;
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = 1;

		double TF = a + (1 - a)*(tf/tf_max);
		double DF = df;

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 8:
 * A (Augmented) : a + (1 - a)(tf/tf_max)
 * T (IDF)       : log(N/df)
 ******************************************************************************/
void TfDf::augmented_idf()
{
	double tf_max = -99999;

	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		if(abow_core->a_ngram_hist_[begin+iter] > tf_max)
			tf_max = abow_core->a_ngram_hist_[begin+iter];
	}

	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double a = 0.5;
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = a + (1 - a)*(tf/tf_max);
		double DF = log(abow_core->no_of_tracks_/df);

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 9:
 * A (Augmented) : a + (1 - a)(tf/tf_max)
 * P (Prob IDF) : max{0, log[(N - df)/df]}
 ******************************************************************************/
void TfDf::augmented_probidf()
{
	double tf_max = -999;

	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		if(abow_core->a_ngram_hist_[begin+iter] > tf_max)
			tf_max = abow_core->a_ngram_hist_[begin+iter];
	}

	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double a = 0.5;
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = a + (1 - a)*(tf/tf_max);
		double DF = std::max(0.0, log((abow_core->no_of_tracks_ - df)/df));

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 10:
 * B (Boolean) : 1 if tf > 0, 0 otherwise
 * N (No)      : 1
 ******************************************************************************/
void TfDf::boolean_no()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = 1;

		double TF = 0.0; if(tf > 0) { TF = 1; }
		double DF = df;

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 11: 
 * B (Boolean) : 1 if tf > 0, 0 otherwise
 * T (IDF)     : log(N/df)
 ******************************************************************************/
void TfDf::boolean_idf()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = 0.0; if(tf > 0) { TF = 1; }
		double DF = log(abow_core->no_of_tracks_/df);

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 12: 
 * B (Boolean)  : 1 if tf > 0, 0 otherwise
 * P (Prob IDF) : max{0, log[(N - df)/df]}
 ******************************************************************************/
void TfDf::boolean_probidf()
{
	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = 0.0; if(tf > 0) { TF = 1; }
		double DF = std::max(0.0, log((abow_core->no_of_tracks_ - df)/df));

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 13:
 * L (Log Avg) : (1 + log(tf))/(1 + log(avg(tf)))
 * N (No)      : 1
 ******************************************************************************/
void TfDf::logavg_no()
{
	double tf_avg = 0.0;
	double count = 0;

	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		tf_avg += abow_core->a_ngram_hist_[begin+iter];
		count += 1;
	}

	tf_avg = tf_avg/count;

	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = 1;

		double TF = 0.0; if(tf > 0) { TF = (1 + log(tf))/(1 + log(tf_avg)); };
		double DF = df;

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 14: 
 * L (Log Avg) : (1 + log(tf))/(1 + log(avg(tf)))
 * T (IDF)     : log(N/df)
 ******************************************************************************/
void TfDf::logavg_idf()
{
	double tf_avg = 0.0;
	double count = 0;

	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		tf_avg += abow_core->a_ngram_hist_[begin+iter];
		count += 1;
	}

	tf_avg = tf_avg/count;

	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = 0.0; if(tf > 0) { TF = (1 + log(tf))/(1 + log(tf_avg)); };
		double DF = log(abow_core->no_of_tracks_/df);

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

/*******************************************************************************
 * Method 15:
 * L (Log Avg)  : (1 + log(tf))/(1 + log(avg(tf)))
 * P (Prob IDF) : max{0, log[(N - df)/df]}
 ******************************************************************************/
void TfDf::logavg_probidf()
{
	double tf_avg = 0.0;
	double count = 0;

	vector<string>::size_type begin = abow_core->no_of_bins_ * track_index;
	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		tf_avg += abow_core->a_ngram_hist_[begin+iter];
		count += 1;
	}

	tf_avg = tf_avg/count;

	for(vector<string>::size_type iter = 0;
		iter < abow_core->no_of_bins_; ++iter)
	{
		double tf = abow_core->a_ngram_hist_[begin+iter];
		double df = abow_core->m_ngram_hist_across_.find(
							abow_core->v_discovered_words_[iter])->second;

		double TF = 0.0; if(tf > 0) { TF = (1 + log(tf))/(1 + log(tf_avg)); };
		double DF = std::max(0.0, log((abow_core->no_of_tracks_ - df)/df));

		abow_core->a_ngram_hist_[begin+iter] = TF * DF;
	}
}

}
