// Copyright (c) 2013 Vinay Bettadapura (Georgia Tech)

#include <iostream>

#include <cuda.h>

using std::cout;

__device__ float ComputeVectorMagnitudeCUDA(float* hist_array, unsigned i,
											unsigned no_of_hist_bins)
{
	float mag = 0.0;

	unsigned begin = no_of_hist_bins * i;

	for(unsigned i = 0; i < no_of_hist_bins; ++i)
	{
		mag += pow(hist_array[begin + i], 2);
	}
	mag = sqrt(mag);
	
	return mag;
}

__device__ float ComputeDotProductCUDA(float* hist_array, unsigned i,
									unsigned j, unsigned no_of_hist_bins)
{
	unsigned begin_i = no_of_hist_bins * i;
	unsigned begin_j = no_of_hist_bins * j;
	
	float result = 0.0;
	for(unsigned iter = 0; iter < no_of_hist_bins; ++iter)
	{
		result += hist_array[begin_i + iter] * hist_array[begin_j + iter];
	}

	return result;
}
	
__device__ float ComputeSimilariyUsingVSMMethodCUDA(float* hist_array,
							unsigned i, unsigned j, unsigned no_of_hist_bins)
{
	float mag_a = 0.0, mag_b = 0.0, result = 0.0;

	mag_a = ComputeVectorMagnitudeCUDA(hist_array, i, no_of_hist_bins);
	mag_b = ComputeVectorMagnitudeCUDA(hist_array, j, no_of_hist_bins);

	if(mag_a == 0 || mag_b == 0)
		return 0;

	result = ComputeDotProductCUDA(hist_array, i, j, no_of_hist_bins);

	return result/(mag_a * mag_b);
}

__global__ void VectorSpaceSimilarityKernel(float* hist_array, float* matrix,
								unsigned no_of_tracks, unsigned no_of_hist_bins)
{
	int i = blockIdx.x * blockDim.x + threadIdx.x;
	int j = blockIdx.y * blockDim.y + threadIdx.y;
	
	if(i >= no_of_tracks || j >= no_of_tracks)
		return;

	float aff = 0.0;

	if(j == i)
	{
		aff = 1.0;
	}
	else if(j > i)
	{
		aff = ComputeSimilariyUsingVSMMethodCUDA(hist_array, i, j,
												no_of_hist_bins);
	}
	else
	{
		return;
	}

	matrix[i * no_of_tracks + j] = aff;
	matrix[j * no_of_tracks + i] = aff;
}

namespace core {

int VectorSpaceSimilarityCUDA(float* a_ngram_hist_, float* a_ngram_hist_device_,
							unsigned long a_ngram_hist_bytes_,
							float* a_affinity_, float* a_affinity_device_,
							unsigned long a_affinity_bytes_,
							size_t no_of_tracks_, size_t no_of_bins_)
{
	cudaError_t error = cudaMalloc((void**)&a_ngram_hist_device_,
								a_ngram_hist_bytes_);
	if(error != cudaSuccess || a_ngram_hist_device_ == 0)
	{
		cout << "\nMemory allocation failed on GPU: "
			<< cudaGetErrorString(error) << "\n";
		return -1;
	}
	cout << "\nBuilding augmented bag-of-words: Allocated memory successfully"
			" on GPU\n";

	error = cudaMalloc((void**)&a_affinity_device_, a_affinity_bytes_);
	if(error != cudaSuccess || a_affinity_device_ == 0)
	{
		cout << "Memory allocation failed on GPU: "
			<< cudaGetErrorString(error) << "\n";
		return -1;
	}
	cout << "Building affinity Matrix: Allocated memory successfully on GPU\n";

	error = cudaMemcpy(a_ngram_hist_device_,
					a_ngram_hist_,
					a_ngram_hist_bytes_,
					cudaMemcpyHostToDevice);
	if(error != cudaSuccess)
	{
		cout << "CUDA memcpy copy from host to device failed: "
			<< cudaGetErrorString(error) << "\n";
		return -1;
	}

	dim3 block_size;
	block_size.x = 32;
	block_size.y = 32;

	dim3 grid_size;
	grid_size.x = (no_of_tracks_ + block_size.x - 1)/block_size.x;
	grid_size.y = (no_of_tracks_ + block_size.y - 1)/block_size.y;

	VectorSpaceSimilarityKernel<<<grid_size, block_size>>>(a_ngram_hist_device_,
								a_affinity_device_, no_of_tracks_, no_of_bins_);

	error = cudaGetLastError();
	if(error != cudaSuccess)
	{
		cout << "CUDA kernel call failed: " << cudaGetErrorString(error) << "\n";
		return -1;
	}

	error = cudaThreadSynchronize();
	if(error != cudaSuccess)
	{
		cout << "CUDA kernel call failed: " << cudaGetErrorString(error) << "\n";
		return -1;
	}

	error = cudaMemcpy(a_affinity_,
					a_affinity_device_,
					a_affinity_bytes_,
					cudaMemcpyDeviceToHost);
	if(error != cudaSuccess)
	{
		cout << "CUDA memory copy from device to host failed: "
			<< cudaGetErrorString(error) << "\n";
		return -1;
	}

	cudaFree(a_ngram_hist_device_);
	cudaFree(a_affinity_device_);

	return 0;
}

}
