// Copyright (c) 2013 Vinay Bettadapura (Georgia Tech)

#include <iostream>

#include "abow_core.h"

#include <boost/program_options/cmdline.hpp>
#include <boost/program_options/options_description.hpp>
#include <boost/program_options/parsers.hpp>
#include <boost/program_options/positional_options.hpp>
#include <boost/program_options/variables_map.hpp>

using std::cerr;
using std::cout;
using std::endl;
using std::string;

namespace po = boost::program_options;

bool process_command_line(
		int argc, char* argv[],
		string& data_source,
		int& ngram_length,
		int& no_of_time_bins,
		int& scoring_metric_id,
		string& encoding,
		int& time_precision,
		string& reg_exp,
		string& duration_bins,
		string& time_unit,
		string& pyramid,
		int& num_of_random_regexp,
		int& similarity_precision,
		string& print_zero_scores)
{
	try
	{
		po::options_description generic("Generic options");
		generic.add_options()
			("help", "Produce help message")
		;

		po::options_description required("Required options");
		required.add_options()
			("data-source,d",
				po::value<string>(&data_source)->required(),
				"Path to the data-source. Example: --data-source "
				"../data")
			("ngram-length,n",
				po::value<int>(&ngram_length)->required(),
				"Length of the ngrams. Example: --ngram-length 3")
			("num-time-bins,N",
				po::value<int>(&no_of_time_bins)->required(),
				"Number of time bins. Example: --num-time-bins 5")
			("scoring-metric-id,s",
				po::value<int>(&scoring_metric_id)->required(),
				"Term scoring metric ID, between 1 and 15. Example: "
				"--scoring-metric-id 3")
			("encoding,e",
				po::value<string>(&encoding)->required(),
				"Encoding scheme; options are 'cumulative', 'interspersed' and "
				"'none'. Example: --encoding cumulative")
		;

		po::options_description optional("Optional options");
		optional.add_options()
			("time-precision,t",
				po::value<int>(&time_precision)->default_value(8),
				"Number of decimal points to consider when quantizing time. "
				"Example: --time-precision 6")
			("rex-exp,r",
				po::value<string>(&reg_exp)->default_value(""),
				"For inserting (space delimited) regular expressions (for "
				"inserting domain knowledge). Example: --reg-exp "
				"\"^0.* A[B|C]A\"")
			("duration-bins,b",
				po::value<string>(&duration_bins)->default_value(""),
				"Set the time quantization bin boundaries manually. Example: "
				"--duration-bins \"2 4 8 16\"")
			("time-unit,u",
				po::value<string>(&time_unit)->default_value("seconds"),
				"Set time measurement unit; options are 'seconds' and 'frames'."
				" Example: --time-unit seconds")
			("pyramid-ngrams,p",
				po::value<string>(&pyramid)->default_value("no"),
				"Use pyramid n-grams; options are 'yes' and 'no'. Example: "
				"--pyramid-ngrams no")
			("random-regexps,R",
				po::value<int>(&num_of_random_regexp)->default_value(0),
				"Number of random regular expressions to use, greater than 0. "
				"Example: --random-regexps 100")
			("similarity-precision,S",
				po::value<int>(&similarity_precision)->default_value(6), 
				"Precision to which to print similarity score values in the "
				"similarity output file. Example: --similarity-precision 3")
			("print-zero-scores,z",
				po::value<string>(&print_zero_scores)->default_value("yes"),
				"Whether to print (yes) or omit (no) activities with 0"
				" similarity scores in the similarity file. Example: "
				"--print-zero-scores no")
		;

		po::options_description cmdline_options;
		cmdline_options.add(generic).add(required).add(optional);

		po::variables_map vm;
		po::store(po::parse_command_line(argc, argv, cmdline_options), vm);

		if(vm.count("help"))
		{
			cout << cmdline_options << "\n";
			return false;
		}

		po::notify(vm);
	}
	catch(std::exception& e)
	{
		std::cerr << "Error: " << e.what() << "\n";
		std::cerr << "See --help for details\n";
		return false;
	}
	catch(...)
	{
		std::cerr << "Unknown error!" << "\n";
		return false;
	}
	
	if(scoring_metric_id < 1 || scoring_metric_id > 16)
	{
		cerr << "Invalid option for --scoring-metric-id\n"
				"See --help for details\n";
		return false;
	}
	
	if(encoding != "cumulative" &&
		encoding != "interspersed" &&
	 	encoding != "none")
	{
		cerr << "Invalid option for --encoding\nSee --help for details\n";
		return false;
	}
	
	if(time_unit != "seconds" && time_unit != "frames")
	{
		cerr << "Invalid option for --time-unit\nSee --help for details\n";
		return false;
	}
	
	if(pyramid != "yes" && pyramid != "no")
	{
		cerr << "Invalid option for --pyramid-ngrams\nSee --help for details\n";
		return false;
	}
	
	if(num_of_random_regexp < 0)
	{
		cerr << "Invalid option for --random-regexps\nSee --help for details\n";
		return false;
	}

	if(similarity_precision < 0) 
	{
		cerr << "Invalid option for --matchfile-precision\n"
				"See --help for details\n";
		return false;
	}

	if(print_zero_scores != "yes" && print_zero_scores != "no")
	{
		cerr << "Invalid option for --print-zero-scores\n"
				"See --help for details\n";
		return false;
	}

	return true;
}

int main(int argc, char *argv[])
{
	string data_source;
	string encoding;
	string reg_exp;
	string duration_bins;
	string time_unit;
	string pyramid;
	string print_zero_scores;

	int ngram_length;
	int no_of_time_bins;
	int scoring_metric_id;
	int time_precision;
	int num_of_random_regexp;
	int similarity_precision;

	bool result = process_command_line(argc, argv, data_source, ngram_length,
						no_of_time_bins, scoring_metric_id, encoding,
						time_precision, reg_exp, duration_bins, time_unit,
						pyramid, num_of_random_regexp, similarity_precision,
						print_zero_scores);

	if(!result)
		return -1;

	cout << "\n";
	cout << "Using data-source: " << data_source << "\n";
	cout << "N-gram length, n = " << ngram_length << "\n";
	cout << "Number of time-bins, N = " << no_of_time_bins << "\n";
	cout << "Scoring metric used: #" << scoring_metric_id << "\n";
	cout << "Encoding scheme: " << encoding << "\n";
	cout << "Time quantization precision: " << time_precision
		<< " decimal places\n";
	if(reg_exp != "")
	{
		cout << "Regular expression: " << reg_exp << "\n";
	}
	if(duration_bins != "")
	{
		cout << "Duration bins: " << duration_bins << "\n";
	}
	cout << "Time measured in: " << time_unit << "\n";
	cout << "Using pyramid encoding: " << pyramid << "\n";
	cout << "Number of random regular expressions used: "
		 << num_of_random_regexp << "\n";
	cout << "Writing to similarity file with precision: "
		<< similarity_precision << " decimal places\n";
	cout << "Print zero scores: " << print_zero_scores << "\n";

	core::ABOWCore abow_core;

	abow_core.set_data_source(data_source);
	abow_core.set_ngram_length(ngram_length);
	abow_core.set_no_of_time_bins(no_of_time_bins);
	abow_core.set_scoring_metric_id(scoring_metric_id);
	abow_core.set_encoding(encoding);
	abow_core.set_time_precision(time_precision);
	abow_core.set_regular_expressions(reg_exp);
	abow_core.set_duration_bins(duration_bins);
	abow_core.set_time_unit(time_unit);
	abow_core.set_pyramid(pyramid);
	abow_core.set_num_of_random_regexp(num_of_random_regexp);
	abow_core.set_similarity_precision(similarity_precision);
	abow_core.set_print_zero_scores(print_zero_scores);

	if(abow_core.Compute() < 0)
	{
		cerr << "Error in computation" << endl;
		return -1;
	}

	return 0;
}
